
function TraceAPI( funcname, ...)
	local msg = string.format("call: %s", funcname )
	for i, v in ipairs( arg ) do
		msg = msg.." "..tostring( arg[i] )
	end
	print( msg )
end



IDLE_ST = 0
FOLLOW_ST = 1
CHASE_ST = 2
ATTACK_ST = 3
MOVE_CMD_ST = 4
STOP_CMD_ST = 5
ATTACK_OBJECT_CMD_ST = 6
ATTACK_AREA_CMD_ST = 7
PATROL_CMD_ST = 8
HOLD_CMD_ST = 9 
SKILL_OBJECT_CMD_ST = 10
SKILL_AREA_CMD_ST = 11
FOLLOW_CMD_ST = 12


function MoveToOwner( id )
	TraceAPI("MoveToOwner", id )
	
	local map = Map.getForegroundMap()
	local obj = map:getObject( id )
	map:moveToSight( id, obj.ownerid, 1 )
end

function Move( id, x, y )
	TraceAPI("Move", id, x, y )
	
	Map.getForegroundMap():moveObjectAbs( id, x, y )
end

function TryAttack( obj )
	assert( obj.targetid )
	local atkrange = obj.atkrange or 1	-- ftHg1
	if GetDistance2( obj.id, obj.targetid ) <= atkrange then	-- U͈͓̎
		obj.motion = MOTION_ATTACK2
	else
		Map.getForegroundMap():moveToSight( obj.id, obj.targetid, atkrange )
	end
end

function Attack( id1, id2 )
	TraceAPI("Attack")
	local map = Map.getForegroundMap()
	
	assert( map:isActor( id1 ) )
	assert( not IsOutOfSight( id1, id2 ) and IsInAttackSight( id1, id2 ) )	-- Ώۂ܂ł̐ڋ߂̓zAIɂ܂
	
	if map:getObject( id2 ).motion == MOTION_DEAD then
		print("error: Attack() target already dead", id2 )
		return	-- łɎłꍇ
	end
	local obj = map:getObject( id1 )
	obj.targetid = id2
	TryAttack( obj )
end

function GetV( v, id )
	TraceAPI("GetV", v, id )
	
	local map = Map.getForegroundMap()
	local obj = map:getObject( id )
	if not obj then
		return -1
	end
	if v == V_OWNER then
		return obj.ownerid
	elseif v == V_POSITION then
		return obj.x - 1, obj.y - 1	-- zAIł̓[_ŌvZĂۂH
	elseif v == V_TYPE then
		-- Ȃ H
	elseif v == V_MOTION then
		return obj.motion
	elseif v == V_ATTACKRANGE then
		return 1 -- ߂
	elseif v == V_TARGET then	-- G[0ۂ
		if( not obj.targetid or not map:isValidObjectId( obj.targetid ) ) then	-- IDw肳Ă鎞
			return 0
		end
		return obj.targetid or 0
	elseif v == V_SKILLATTACKRANGE then
		return 1 -- ߂Q
	elseif v == V_HOMUNTYPE then
		return obj.homtype
	elseif v == V_HP then
		return obj.HP
	elseif v == V_SP then
		return obj.SP
	elseif v == V_MAXHP then
		return obj.MHP
	elseif v == V_MAXSP then
		return obj.MSP
	else
		assert( nil )
	end
end

function GetActors()
	TraceAPI("GetActors")
	local map = Map.getForegroundMap()
	for k, v in map:enumObjects() do
		if v.type == MAPOBJ_I then
			return map:getSightActors( k )
		end
	end
	
	assert( nil )
end

function GetTick()
	TraceAPI("GetTick")
	return 100
end

function GetMsg( id )
	TraceAPI("GetMsg", id )
	return { NONE_CMD }
end

function GetResMsg( id )
	TraceAPI("GetResMsg", id )
	return { NONE_CMD }
end

function SkillObject( id, level, skill, target )
	TraceAPI("SkillObject", id, level, skill, target )
end

function SkillGround( id, level, skill, x, y )
	TraceAPI("SkillGround", id, level, skill, x, y )
end

function TraceAI( str )
	print( string.format("trace: %s", str ) )
end

function IsMonster( id )	-- AI.lua  1 or 0 ԂۂH
	TraceAPI("IsMonster", id )

	if not Map.getForegroundMap():isValidObjectId( id ) then
		return 0
	end
	if Map.getForegroundMap():getObject( id ).type == MAPOBJ_ENEMY then
		return 1
	else
		return 0
	end
end
