
MAP_WIDTH = 20
MAP_HEIGHT = 30

MAPOBJ_NULL = 0
MAPOBJ_OWNER = 1
MAPOBJ_I = 2
MAPOBJ_HOMUN = 3
MAPOBJ_NPC = 4
MAPOBJ_PET = 5
MAPOBJ_ENEMY = 6
MAPOBJ_PC = 7

--[[
Stack = {}

function Stack.push( v )
	local len = table.getn( self )
	self[ len + 1 ] = v
	table.setn( self, len + 1 )
end

function Stack.pop()
end

function Stack.front()
end

function Stack.back()
end
--]]

Map = {}
-- Map.Object { id, type, motion, x, y, destx, desty, targetid, atkrange }
-- Map.Object.Status { skilleffect, MHP, MSP, HP, SP, ATK }
-- Map.Homun { homtype, ownerid }
-- Map.Enemy { ai }

function NewStatus( mhp, msp )
	return { MHP = mhp, HP = mhp, MSP = msp, SP = msp }
end

function Map.setForegroundMap( map )
	Map.fginst = map
end

function Map.getForegroundMap()
	return Map.fginst
end

function Map.nullObject()
	return { id = 0, type = MAPOBJ_NULL }
end

function Map.new()
	local map = {}
	setmetatable( map, { __index = Map } )
	map.lines = {}
	for i = 1, MAP_HEIGHT, 1 do
		map.lines[i] = { cells = {} }
		for j = 1, MAP_WIDTH, 1 do
			map.lines[i].cells[j] = { stack = { Map.nullObject(), n = 1 } }
		end
	end
	map.objtable = {}
	return map
end

function Map:objectToChar( obj )
	local t = {};
	t[ MAPOBJ_NULL ] = ""
	t[ MAPOBJ_OWNER]  = ""
	t[ MAPOBJ_I ] = ""
	t[ MAPOBJ_HOMUN ] = "z"
	t[ MAPOBJ_NPC ] = "m"
	t[ MAPOBJ_PET ] = "y"
	t[ MAPOBJ_ENEMY ] = "G"
	t[ MAPOBJ_PC ] = "l"

	if obj.motion == MOTION_DEAD then
		assert( obj.HP and obj.HP == 0 )
		return ""
	end
	return t[ obj.type ]
end

function Map:printMap()
	for i = 1, MAP_HEIGHT, 1 do
		local line = ""
		for j = 1, MAP_WIDTH, 1 do
			line = line.." "..self:objectToChar( self.lines[i].cells[j].stack[1] )
		end
		print( line )
	end
end

function Map:insertObject( obj, x, y )
	assert( self:isValidObjectId( obj.id ) )
	assert( self:isCell( x, y ) )
	
	local list = self.lines[y].cells[x].stack
--	if list[1].type ~= MAPOBJ_NULL then	-- kIuWFNg}[J[ƂĎcĂƂ
		for i = list.n, 1, -1 do
			list[i+1] = list[i]
		end
		list.n = list.n + 1
--	end
	list[1] = obj
	obj.x = x
	obj.y = y
	
	print( string.format("insert object: id=%d type=%s x=%d y=%d n=%d",
		obj.id, self:objectToChar( obj ), x, y, list.n ) )
	return obj
end

function Map:setObjectStatus( id, stat )
	assert( self:isValidObjectId( id ) )
	
	local obj = self:getObject( id )
	for k, v in pairs( stat ) do	-- Deep Copy
		obj[ k ] = v
	end
	return obj
end

function Map:createObject( id, type, stat )
	local obj = { id = id, type = type, motion = MOTION_STAND }
	self.objtable[ id ] = obj
	if stat then
		self:setObjectStatus( id, stat )
	end
	return obj
end

function Map:getObject( id )
	return self.objtable[ id ]
end

function Map:enumObjects()
	return pairs( self.objtable )
end

function Map:getCellObjects( x, y )
	return self.lines[y].cells[x].stack
end

function Map:spawnObject( obj, x, y, ai )
	if ai then
		obj.ai = ai
	end
	return self:insertObject( obj, x, y )
end

function Map:isCell( x, y )
	return x >= 1 and x <= MAP_WIDTH
		and y >= 1 and y <= MAP_HEIGHT
end

function Map:moveObject( id, x, y )	-- ړ x, y ZĂ
	assert( self:isActor( id ) )
	print("move:", id, x, y )
	local obj = self:getObject( id )
	local nx, ny
	if obj.motion ~= MOTION_MOVE then
		nx = obj.x + x
		ny = obj.y + y
		if not self:isCell( nx, ny ) then return end
		obj.motion = MOTION_MOVE
	else
		nx = obj.destx + x
		ny = obj.desty + y
	end
	if self:isCell( nx, ny ) then
		obj.destx = nx
		obj.desty = ny
	end
end

function Map:moveObjectAbs( id, x, y )	-- ΈʒuɈړ
	assert( self:isActor( id ) )
	print("move abs:", id, x, y )
	if not self:isCell( x, y ) then return end
	local obj = self:getObject( id )
	obj.motion = MOTION_MOVE
	obj.destx = x
	obj.desty = y
end

--[[
function Map:killObject( id )
	self.getObject( id ).motion = MOTION_DEAD
	-- typeɂĂ͎stepMotionŃIuWFNgerase
end
--]]

function Map:eraseObject( id )	-- IuWFNg}bvBi폜͂Ȃj
	local obj = self:getObject( id )
	local stack = self.lines[obj.y].cells[obj.x].stack
	for i, v in ipairs( stack ) do
		if v == obj then
			for i = i, stack.n, 1 do
				stack[i] = stack[i+1]
			end
			stack.n = stack.n - 1
		end
	end
	obj.x = 0	-- WNA
	obj.y = 0
	
	obj = stack[1]
	print( string.format("erase object: result id=%d type=%s x=%d y=%d n=%d",
		obj.id, self:objectToChar( obj ), obj.x or -1, obj.y or -1, stack.n ) )	-- MAPOBJ_NULL  .x .y == nil
end

function Map:destroyObject( id )
	table.remove( self.objtable, id )
end

function Map:isValidObjectId( id )
	return id and id > 0 and not not self:getObject( id )
end

function Map:isActor( id )	-- ΏۂMAPɂĎłȂ
	local obj = self:getObject( id )
	return self:isValidObjectId( id ) and self:isCell( obj.x, obj.y ) and obj.motion ~= MOTION_DEAD
end

function MinDistance( n1, n2 )
	local dis = n1 - n2
	if math.abs( dis ) <= 1 then	-- n ̂ꂪ1Zȓ̏ꍇ
		return 0
	else
	--	return dis
		if dis > 0 then
			return dis - 1
		else
			return dis + 1
		end
	end
end

function Map:moveToSight( id, targetid, sight )	-- attack sight ͖ł߂Ȃ
	local obj = self:getObject( id )
	local target = self:getObject( targetid )
	local nx = obj.x - MinDistance( obj.x, target.x )
	local ny = obj.y - MinDistance( obj.y, target.y )
	assert( self:isCell( nx, ny ) )
	self:moveObjectAbs( id, nx, ny )
end

function Map:getSightActors( id )
	local obj = self:getObject( id )
	local list = {}
	local size = 0
	for k, v in Map.getForegroundMap():enumObjects() do
		if k ~= id and v.motion ~= MOTION_DEAD and not IsOutOfSight( id, k ) then	-- ̂͏O
			size = size + 1
			list[ size ] = k
			list.n = size
			
			print( string.format("visible actor: %d,%d -> %d,%d", obj.x, obj.y, v.x, v.y ) )
		end
	end
	return list
end

function IsEnemy( type1, type2 )
	if type1 == MAPOBJ_ENEMY then
		return type2 == MAPOBJ_I or type2 == MAPOBJ_OWNER or type2 == MAPOBJ_PC
	else
		return IsMonster( type2 ) == 1
	end
end

function Map:getNearestEnemy( id )	-- ς
	local obj = self:getObject( id )
	if obj.type == MAPOBJ_ENEMY then
		local enemys = {}
		for k,v in ipairs( self:getSightActors( id ) ) do
			if IsEnemy( self:getObject( id ).type, self:getObject( v ).type ) then
				enemys[ table.getn( enemys ) + 1 ] = v
			end
		end
		local min_dis = 100
		local dis
		local result = 0
		for i, v in ipairs( enemys ) do
			dis = GetDistance2( id, v )
			if dis < min_dis then
				result = v
				min_dis = dis
			end
		end
		return result
	else
		-- (P~I[g^Q)p GetMyEnemyB
	end
	return nil
end

function Map:stepMotion()
	for k, obj in self:enumObjects() do
		if self:isCell( obj.x, obj.y ) then	-- ̃IuWFNĝ
			if obj.ai then	-- AIĂ
				obj.ai:motion( self, obj )
			end
		end
	end
end


HELP_MESSAGE = [[
		
		ւՂł
		------------
		w 
		a 
		s 
		d 
		z 
		v 傤ق
		q ܂
		h 
		------------
		
]]
